import json

from NEMO.decorators import staff_member_required
from django.contrib.auth.decorators import login_required
from django.http import StreamingHttpResponse
from django.shortcuts import get_object_or_404, redirect, render

from NEMO_floor_plans.forms import FloorPlanForm, FloorPlanNodeForm
from NEMO_floor_plans.models import FloorPlan, FloorPlanIcon, FloorPlanNode
from NEMO_sensors.models import Sensor
from NEMO.views.pagination import SortedPaginator


def floor_plan_list_page(request):
    name_search = request.GET.get("search")
    floor_plans_list = FloorPlan.objects.filter(name__icontains=name_search) if name_search else FloorPlan.objects.all()
    return SortedPaginator(floor_plans_list, request, order_by="-name").get_current_page()


@login_required
def floor_plans(request, floor_plan_id=None):
    if request.method == "POST":
        # If action is to set or remove floor plan as default
        if "set_default" in request.POST or "remove_default" in request.POST:
            target_floor_plan_id = (
                request.POST["set_default"] if "set_default" in request.POST else request.POST["remove_default"]
            )
            target_floor_plan = get_object_or_404(FloorPlan, id=target_floor_plan_id)
            target_floor_plan.is_default = "set_default" in request.POST
            target_floor_plan.save()
            return redirect("floor_plans")

    loaded_floor_plan = (
        FloorPlan.objects.get(id=floor_plan_id) if floor_plan_id else FloorPlan.objects.filter(is_default=True).first()
    )
    page = floor_plan_list_page(request)
    node_list = [node.serialize(False) for node in (loaded_floor_plan.nodes.all() if loaded_floor_plan else [])]

    return render(
        request,
        "floor_plans.html",
        {
            "page": page,
            "paginator": page.paginator,
            "search": request.GET.get("search", ""),
            "loaded_floor_plan": loaded_floor_plan,
            "node_list": node_list,
            "is_staff": request.user.is_staff,
        },
    )


def floor_plan_node_save(request, floor_plan):
    floor_plan_node_id = request.POST.get("id")
    form_data = request.POST.copy()
    form_data["floor_plan"] = floor_plan.id

    if floor_plan_node_id:
        floor_plan_node = FloorPlanNode.objects.get(id=floor_plan_node_id, floor_plan=floor_plan)
        floor_plan_node_form = FloorPlanNodeForm(form_data, instance=floor_plan_node)
    else:
        floor_plan_node_form = FloorPlanNodeForm(form_data)

    if floor_plan_node_form.is_valid():
        floor_plan_node = floor_plan_node_form.save()
        floor_plan_node.save()
        return True, floor_plan_node.id, None
    return False, floor_plan_node_form.instance.id, floor_plan_node_form


@login_required
@staff_member_required
def floor_plan_edit(request, floor_plan_id):
    selected_node_id = None
    node_form = None
    floor_plan_form = None
    loaded_floor_plan = get_object_or_404(FloorPlan, id=floor_plan_id)
    if request.method == "POST":
        # If action is to delete a node
        if "delete_node" in request.POST:
            FloorPlanNode.objects.get(id=request.POST["id"], floor_plan__id=floor_plan_id).delete()
            return redirect("floor_plan_edit", floor_plan_id=floor_plan_id)

        # If action is to save/create a node
        if "save_node" in request.POST:
            save_successful, selected_node_id, node_form = floor_plan_node_save(request, loaded_floor_plan)
            if save_successful:
                return redirect("floor_plan_edit", floor_plan_id=floor_plan_id)

        # If action is to save the floor plan
        if "save_floor_plan" in request.POST:
            # object get or 404
            floor_plan_form = FloorPlanForm(request.POST, request.FILES, instance=loaded_floor_plan)
            if floor_plan_form.is_valid():
                floor_plan_form.save()
                return redirect("floor_plan_edit", floor_plan_id=floor_plan_id)

    node_list = [node.serialize(False) for node in loaded_floor_plan.nodes.all()]
    icons = FloorPlanIcon.objects.all()
    types = FloorPlanNode.NodeType.Choices
    sensors = Sensor.objects.all().order_by("name")
    sensors_with_nodes = sensors.filter(floorplannode__isnull=False).distinct()
    sensors_without_nodes = sensors.filter(floorplannode__isnull=True).distinct()

    return render(
        request,
        "floor_plan_editor.html",
        {
            "loaded_floor_plan": loaded_floor_plan,
            "node_list": node_list,
            "icons": icons,
            "types": types,
            "sensors_with_nodes": sensors_with_nodes,
            "sensors_without_nodes": sensors_without_nodes,
            "selected_node_id": selected_node_id,
            "node_form": node_form,
            "floor_plan_form": floor_plan_form,
            "floor_plans": FloorPlan.objects.all(),
        },
    )


@login_required
def floor_plan_nodes(request, floor_plan_id):
    floor_plan = get_object_or_404(FloorPlan, id=floor_plan_id)
    separator = "\x1e"
    nodes = floor_plan.nodes.filter(type__exact=FloorPlanNode.NodeType.SENSOR)

    def stream():
        for node in nodes:
            yield json.dumps(node.serialize()) + separator

    return StreamingHttpResponse(stream(), content_type="application/json")



import os

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

DEBUG = True
DEFAULT_AUTO_FIELD = "django.db.models.AutoField"
FIXTURE_DIRS = ["/nemo/"]
AUTH_USER_MODEL = "NEMO.User"
WSGI_APPLICATION = "NEMO.wsgi.application"
ROOT_URLCONF = "NEMO.urls"

LOGIN_URL = "login"
LOGIN_REDIRECT_URL = "login"

DATETIME_FORMAT = "l, F jS, Y @ g:i A"
DATE_FORMAT = "m/d/Y"
TIME_FORMAT = "g:i A"
DATETIME_INPUT_FORMATS = ["%m/%d/%Y %I:%M %p"]
DATE_INPUT_FORMATS = ["%m/%d/%Y"]
TIME_INPUT_FORMATS = ["%I:%M %p"]

USE_I18N = False
USE_L10N = False
USE_TZ = True

INSTALLED_APPS = [
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "django.contrib.admin",
    "django.contrib.humanize",
    "NEMO_floor_plans",
    "NEMO_sensors",
    "NEMO",
    "NEMO.apps.kiosk",
    "NEMO.apps.area_access",
    "rest_framework",
    "django_filters",
    "mptt",
]

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
    "django.middleware.common.BrokenLinkEmailsMiddleware",
    "NEMO.middleware.DeviceDetectionMiddleware",
    "NEMO.middleware.RemoteUserAuthenticationMiddleware",
    "NEMO.middleware.ImpersonateMiddleware",
]

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "NEMO.context_processors.hide_logout_button",  # Add a 'request context processor' in order to figure out whether to display the logout button. If the site is configured to use the LDAP authentication backend then we want to provide a logoff button (in the menu bar). Otherwise the Kerberos authentication backend is used and no logoff button is necessary.
                "NEMO.context_processors.base_context",  # Base context processor
                "django.contrib.auth.context_processors.auth",
                "django.template.context_processors.debug",
                "django.template.context_processors.media",
                "django.template.context_processors.static",
                "django.template.context_processors.request",
                "django.template.context_processors.tz",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

REST_FRAMEWORK = {
    "DEFAULT_PERMISSION_CLASSES": ("NEMO.permissions.DjangoModelPermissions",),
    "DEFAULT_FILTER_BACKENDS": ("django_filters.rest_framework.DjangoFilterBackend",),
    "DEFAULT_PAGINATION_CLASS": "rest_framework.pagination.PageNumberPagination",
    "PAGE_SIZE": 1000,
}

SERVER_EMAIL = "NEMO Server Administrator <nemo.admin@example.org>"

ADMINS = [
    ("System administrator", "sysadmin@example.org"),
]
MANAGERS = ADMINS

EMAIL_HOST = "mail.example.org"
EMAIL_PORT = 25

EMAIL_FILE_PATH = "./email_logs"

TIME_ZONE = "US/Eastern"

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": "./test_nemo.db",
    }
}

STATIC_URL = "/static/"
MEDIA_ROOT = BASE_DIR + "/../resources/emails"
MEDIA_URL = "/media/"

SECRET_KEY = "test"

ALLOW_CONDITIONAL_URLS = True
AUTHENTICATION_BACKENDS = ["NEMO.views.authentication.NginxKerberosAuthorizationHeaderAuthenticationBackend"]

IDENTITY_SERVICE = {
    "available": False,
    "url": "https://identity.example.org/",
    "domains": [],
}

LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "verbose": {
            "format": "%(levelname)s %(asctime)s %(module)s %(process)d %(thread)d %(message)s",
            "datefmt": "%d/%b/%Y %H:%M:%S",
        },
    },
    "handlers": {
        "file": {
            "level": "DEBUG",
            "class": "logging.FileHandler",
            "filename": "./test_nemo.log",
            "formatter": "verbose",
        },
        "console": {
            "formatter": "verbose",
            "class": "logging.StreamHandler",
        },
    },
    "loggers": {
        "": {
            "handlers": ["file", "console"],
            "level": "DEBUG",
            "propagate": True,
        },
        "django": {
            "handlers": ["file", "console"],
            "level": "DEBUG",
            "propagate": True,
        },
        "django.template": {
            "handlers": ["file", "console"],
            "level": "INFO",
            "propagate": True,
        },
        "NEMO": {
            "level": "DEBUG",
            "handlers": ["file", "console"],
            "propagate": True,
        },
        "NEMO.middleware": {
            "level": "INFO",
            "handlers": ["file", "console"],
            "propagate": True,
        },
    },
}



from django.apps import AppConfig


class FloorPlansConfig(AppConfig):
    name = "NEMO_floor_plans"
    verbose_name = "NEMO Floor Plans"
    default_auto_field = "django.db.models.AutoField"

    def ready(self):
        from NEMO.plugins.utils import check_extra_dependencies

        """
        This code will be run when Django starts.
        """
        check_extra_dependencies(self.name, ["NEMO", "NEMO-CE"])
        check_extra_dependencies(self.name, ["nemo-sensors"])



from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("NEMO_floor_plans", "0002_floorplannode_sensor_text_position"),
    ]

    def update_coordinates(apps, schema_editor):
        FloorPlanNode = apps.get_model("NEMO_floor_plans", "FloorPlanNode")

        for node in FloorPlanNode.objects.all():
            floor_plan = node.floor_plan
            image_width = floor_plan.image.width
            image_height = floor_plan.image.height
            node.x = int(node.x * image_width)
            node.y = int(node.y * image_height)
            node.save()

    def revert_coordinates(apps, schema_editor):
        FloorPlanNode = apps.get_model("NEMO_floor_plans", "FloorPlanNode")

        for node in FloorPlanNode.objects.all():
            floor_plan = node.floor_plan
            image_width = floor_plan.image.width
            image_height = floor_plan.image.height
            node.x = node.x / image_width
            node.y = node.y / image_height
            node.save()

    operations = [
        migrations.RunPython(update_coordinates, revert_coordinates),
        migrations.AlterField(
            model_name="floorplannode",
            name="x",
            field=models.IntegerField(help_text="The x coordinate of the node"),
        ),
        migrations.AlterField(
            model_name="floorplannode",
            name="y",
            field=models.IntegerField(help_text="The y coordinate of the node"),
        ),
    ]



from django.forms import ModelForm

from NEMO_floor_plans.models import FloorPlan, FloorPlanNode


class FloorPlanNodeForm(ModelForm):
    class Meta:
        model = FloorPlanNode
        fields = "__all__"


class FloorPlanForm(ModelForm):
    class Meta:
        model = FloorPlan
        fields = "__all__"



from django.contrib import admin
from NEMO.mixins import ModelAdminRedirectMixin
from django.utils.safestring import mark_safe

from NEMO_floor_plans.models import FloorPlan, FloorPlanIcon, FloorPlanNode


@admin.register(FloorPlan)
class FloorPlanAdmin(ModelAdminRedirectMixin, admin.ModelAdmin):
    list_display = ["name", "is_default", "image_thumbnail"]

    @admin.display(ordering="image", description="Image")
    def image_thumbnail(self, obj):
        if obj.image:
            return mark_safe(f'<img src="{obj.image.url}" height="250px" />')
        return "No image"


@admin.register(FloorPlanNode)
class FloorPlanNodeAdmin(ModelAdminRedirectMixin, admin.ModelAdmin):
    list_display = ["floor_plan", "type", "icon_thumbnail", "sensor", "portal_to"]
    list_filter = ["type", ("icon", admin.RelatedOnlyFieldListFilter)]

    @admin.display(ordering="icon", description="Icon")
    def icon_thumbnail(self, obj: FloorPlanNode):
        if obj.icon:
            return mark_safe(f'<img src="{obj.icon.image.url}" height="50px" />')
        return "No icon"


@admin.register(FloorPlanIcon)
class FloorPlanIconAdmin(ModelAdminRedirectMixin, admin.ModelAdmin):
    list_display = ["name", "image_thumbnail"]

    @admin.display(ordering="image", description="Image")
    def image_thumbnail(self, obj):
        if obj.image:
            return mark_safe(f'<img src="{obj.image.url}" height="50px" />')
        return "No image"



from NEMO.utilities import format_datetime
from django.core.exceptions import ValidationError
from django.db import models
from django.forms import model_to_dict

from NEMO_sensors.models import Sensor
from NEMO.models import BaseModel


class FloorPlan(BaseModel):
    name = models.CharField(max_length=100, blank=False, null=False, help_text="The name of the floor plan")
    image = models.ImageField(
        upload_to="floor_plans/", blank=False, null=False, help_text="The image of the floor plan"
    )
    is_default = models.BooleanField(default=False, help_text="Whether this floor plan is the default one")

    def serialize(self):
        data = model_to_dict(self)
        if self.image:
            data["image"] = self.image.url
        return data

    def __str__(self):
        return self.name

    def save(self, *args, **kwargs):
        if self.pk:
            old_instance = FloorPlan.objects.get(pk=self.pk)

            # delete existing image if it is updated
            if old_instance.image and self.image != old_instance.image:
                old_instance.image.delete(save=False)

            # make sure only one floor plan is default
            if self.is_default and not old_instance.is_default:
                FloorPlan.objects.exclude(pk=self.pk).update(is_default=False)

        super().save(*args, **kwargs)


class FloorPlanIcon(BaseModel):
    name = models.CharField(max_length=100, blank=False, null=False, help_text="The name of the icon")
    image = models.ImageField(upload_to="floor_plan_icons/", blank=False, null=False, help_text="The image of the icon")

    def __str__(self):
        return self.name

    def save(self, *args, **kwargs):
        # delete existing image if it is updated
        if self.pk:
            old_instance = FloorPlanIcon.objects.get(pk=self.pk)
            if old_instance.image and self.image != old_instance.image:
                old_instance.image.delete(save=False)

        super().save(*args, **kwargs)


class FloorPlanNode(BaseModel):
    class NodeType(object):
        PORTAL = 0
        SENSOR = 1
        Choices = ((PORTAL, "Portal"), (SENSOR, "Sensor"))

    class SensorTextPositionType(object):
        TOP = 0
        RIGHT = 1
        BOTTOM = 2
        LEFT = 3
        Choices = ((TOP, "Top"), (RIGHT, "Right"), (BOTTOM, "Bottom"), (LEFT, "Left"))

    floor_plan = models.ForeignKey(FloorPlan, on_delete=models.CASCADE, related_name="nodes")
    type = models.IntegerField(choices=NodeType.Choices, blank=False, null=False, default=NodeType.SENSOR)
    icon = models.ForeignKey(FloorPlanIcon, on_delete=models.SET_NULL, blank=True, null=True)
    icon_scale = models.FloatField(default=1.0)
    x = models.IntegerField(blank=False, null=False, help_text="The x coordinate of the node")
    y = models.IntegerField(blank=False, null=False, help_text="The y coordinate of the node")
    font_size = models.FloatField(default=25, blank=True, null=True, help_text="The font size of the node label")
    sensor_text_position = models.IntegerField(
        choices=SensorTextPositionType.Choices,
        blank=True,
        null=True,
        help_text="The position of the sensor text",
    )
    sensor = models.ForeignKey(Sensor, on_delete=models.SET_NULL, blank=True, null=True)
    portal_to = models.ForeignKey(FloorPlan, on_delete=models.SET_NULL, blank=True, null=True)

    def __str__(self):
        return f"Node ({self.id}) of type {self.type.__str__()} on {self.floor_plan.name}"

    def serialize(self, include_sensor_data=True):
        node_value = {}
        if self.type == FloorPlanNode.NodeType.SENSOR and self.sensor:
            if include_sensor_data:
                node_value["sensor_value"] = self.sensor.last_value_display() if self.sensor.last_value else None
                node_value["sensor_datetime"] = (
                    format_datetime(self.sensor.last_read, "SHORT_DATETIME_FORMAT") if self.sensor.last_read else None
                )
                node_value["sensor_alert"] = self.sensor.alert_triggered()

            node_value["sensor_id"] = self.sensor.id
            node_value["sensor_name"] = self.sensor.name
        if self.type == FloorPlanNode.NodeType.PORTAL and self.portal_to:
            node_value["portal_name"] = self.portal_to.name
            node_value["portal_id"] = self.portal_to.id
            node_value["portal_img"] = self.portal_to.image.url

        return {
            "id": self.id,
            "type": self.get_type_display(),
            "type_id": self.type,
            "icon": self.icon.image.url if self.icon else None,
            "icon_id": self.icon.id if self.icon else None,
            "icon_scale": self.icon_scale,
            "font_size": self.font_size,
            "sensor_text_position": self.sensor_text_position,
            "x": self.x,
            "y": self.y,
            **node_value,
        }

    def clean(self):
        if self.type == self.NodeType.SENSOR:
            if not self.sensor:
                raise ValidationError({"sensor": "Sensor is required."})
            if self.portal_to:
                raise ValidationError({"portal_to": "Sensor node cannot have portal."})
            if self.sensor_text_position is None:
                raise ValidationError({"sensor_text_position": "Sensor text position is required."})
            if not self.font_size:
                raise ValidationError({"font_size": "Font size is required."})

        if self.type == self.NodeType.PORTAL:
            if not self.portal_to:
                raise ValidationError({"portal_to": "Portal is required."})
            if self.sensor:
                raise ValidationError({"type": "Portal node cannot have sensor."})
            if self.floor_plan == self.portal_to:
                raise ValidationError({"portal_to": "Portal cannot be the same as the current floor plan."})



# Generated by Django 4.2.15 on 2024-11-09 18:35

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("sensors", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="FloorPlan",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(help_text="The name of the floor plan", max_length=100)),
                ("image", models.ImageField(help_text="The image of the floor plan", upload_to="floor_plans/")),
                (
                    "is_default",
                    models.BooleanField(default=False, help_text="Whether this floor plan is the default one"),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="FloorPlanIcon",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(help_text="The name of the icon", max_length=100)),
                ("image", models.ImageField(help_text="The image of the icon", upload_to="floor_plan_icons/")),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="FloorPlanNode",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("type", models.IntegerField(choices=[(0, "Portal"), (1, "Sensor")], default=1)),
                ("icon_scale", models.FloatField(default=1.0)),
                ("x", models.FloatField(help_text="The x percentage coordinate of the node")),
                ("y", models.FloatField(help_text="The y percentage coordinate of the node")),
                (
                    "font_size",
                    models.FloatField(blank=True, default=25, help_text="The font size of the node label", null=True),
                ),
                (
                    "floor_plan",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="nodes",
                        to="NEMO_floor_plans.floorplan",
                    ),
                ),
                (
                    "icon",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="NEMO_floor_plans.floorplanicon",
                    ),
                ),
                (
                    "portal_to",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="NEMO_floor_plans.floorplan",
                    ),
                ),
                (
                    "sensor",
                    models.ForeignKey(
                        blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to="sensors.sensor"
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
    ]



# Generated by Django 4.2.15 on 2024-11-10 19:29

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("NEMO_floor_plans", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="floorplannode",
            name="sensor_text_position",
            field=models.IntegerField(
                blank=True,
                choices=[(0, "Top"), (1, "Right"), (2, "Bottom"), (3, "Left")],
                help_text="The position of the sensor text",
                null=True,
            ),
        ),
    ]



from django.urls import path

from NEMO_floor_plans import views

urlpatterns = [
    # Contracts and procurements
    path("floor_plans/", views.floor_plans, name="floor_plans"),
    path("floor_plans/<int:floor_plan_id>/", views.floor_plans, name="floor_plan"),
    path("floor_plans/<int:floor_plan_id>/edit", views.floor_plan_edit, name="floor_plan_edit"),
    path("floor_plans/<int:floor_plan_id>/nodes", views.floor_plan_nodes, name="floor_plan_nodes"),
]
